@@ -26,6 +26,7 @@
 
 mod fs;
 pub mod style;
+mod suffix;
 
 use std::collections::HashMap;
 use std::env;
@@ -33,6 +34,8 @@ use std::ffi::OsString;
 use std::fs::{DirEntry, FileType, Metadata};
 use std::path::{Component, Path, PathBuf, MAIN_SEPARATOR};
 
+use crate::suffix::{SuffixMap, SuffixMapBuilder};
+
 pub use crate::style::{Color, FontStyle, Style};
 
 #[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
@@ -142,8 +145,6 @@ impl Indicator {
     }
 }
 
-type FileNameSuffix = String;
-
 /// Iterator over the path components with their respective style.
 pub struct StyledComponents<'a> {
     /// Reference to the underlying LsColors object
@@ -218,35 +219,91 @@ impl Colorable for DirEntry {
     }
 }
 
+/// Builder for [LsColors].
+struct LsColorsBuilder {
+    indicator_mapping: HashMap<Indicator, Style>,
+
+    /// Whether Indicator::RegularFile falls back to Indicator::Normal
+    /// (see <https://github.com/sharkdp/lscolors/issues/48#issuecomment-1582830387>)
+    file_normal_fallback: bool,
+
+    suffixes: SuffixMapBuilder,
+}
+
+impl LsColorsBuilder {
+    fn empty() -> Self {
+        Self {
+            indicator_mapping: HashMap::new(),
+            file_normal_fallback: true,
+            suffixes: SuffixMapBuilder::default(),
+        }
+    }
+
+    fn add_from_string(&mut self, input: &str) {
+        for entry in input.split(':') {
+            let parts: Vec<_> = entry.split('=').collect();
+
+            if let Some([entry, ansi_style]) = parts.get(0..2) {
+                let style = Style::from_ansi_sequence(ansi_style);
+                if let Some(suffix) = entry.strip_prefix('*') {
+                    self.suffixes.push(suffix, style);
+                } else if let Some(indicator) = Indicator::from(entry) {
+                    if let Some(style) = style {
+                        self.indicator_mapping.insert(indicator, style);
+                    } else {
+                        self.indicator_mapping.remove(&indicator);
+                        if indicator == Indicator::RegularFile {
+                            self.file_normal_fallback = false;
+                        }
+                    }
+                }
+            }
+        }
+    }
+
+    fn build(self) -> LsColors {
+        LsColors {
+            indicator_mapping: self.indicator_mapping,
+            file_normal_fallback: self.file_normal_fallback,
+            suffixes: self.suffixes.build(),
+        }
+    }
+}
+
 const LS_COLORS_DEFAULT: &str = "rs=0:lc=\x1b[:rc=m:cl=\x1b[K:ex=01;32:sg=30;43:su=37;41:di=01;34:st=37;44:ow=34;42:tw=30;42:ln=01;36:bd=01;33:cd=01;33:do=01;35:pi=33:so=01;35:";
 
+impl Default for LsColorsBuilder {
+    fn default() -> Self {
+        let mut builder = Self::empty();
+        builder.add_from_string(LS_COLORS_DEFAULT);
+        builder
+    }
+}
+
 /// Holds information about how different file system entries should be colorized / styled.
 #[derive(Debug, Clone)]
 pub struct LsColors {
     indicator_mapping: HashMap<Indicator, Style>,
 
-    // Note: you might expect to see a `HashMap` for `suffix_mapping` as well, but we need to
-    // preserve the exact order of the mapping in order to be consistent with `ls`.
-    suffix_mapping: Vec<(FileNameSuffix, Option<Style>)>,
+    /// Whether Indicator::RegularFile falls back to Indicator::Normal
+    /// (see <https://github.com/sharkdp/lscolors/issues/48#issuecomment-1582830387>)
+    file_normal_fallback: bool,
+
+    suffixes: SuffixMap,
 }
 
 impl Default for LsColors {
     /// Constructs a default `LsColors` instance with some default styles. See `man dircolors` for
     /// information about the default styles and colors.
     fn default() -> Self {
-        let mut lscolors = LsColors::empty();
-        lscolors.add_from_string(LS_COLORS_DEFAULT);
-        lscolors
+        LsColorsBuilder::default().build()
     }
 }
 
 impl LsColors {
     /// Construct an empty [`LsColors`](struct.LsColors.html) instance with no pre-defined styles.
     pub fn empty() -> Self {
-        LsColors {
-            indicator_mapping: HashMap::new(),
-            suffix_mapping: vec![],
-        }
+        LsColorsBuilder::empty().build()
     }
 
     /// Creates a new [`LsColors`](struct.LsColors.html) instance from the `LS_COLORS` environment
@@ -261,29 +318,9 @@ impl LsColors {
 
     /// Creates a new [`LsColors`](struct.LsColors.html) instance from the given string.
     pub fn from_string(input: &str) -> Self {
-        let mut lscolors = LsColors::default();
-        lscolors.add_from_string(input);
-        lscolors
-    }
-
-    fn add_from_string(&mut self, input: &str) {
-        for entry in input.split(':') {
-            let parts: Vec<_> = entry.split('=').collect();
-
-            if let Some([entry, ansi_style]) = parts.get(0..2) {
-                let style = Style::from_ansi_sequence(ansi_style);
-                if let Some(suffix) = entry.strip_prefix('*') {
-                    self.suffix_mapping
-                        .push((suffix.to_string().to_ascii_lowercase(), style));
-                } else if let Some(indicator) = Indicator::from(entry) {
-                    if let Some(style) = style {
-                        self.indicator_mapping.insert(indicator, style);
-                    } else {
-                        self.indicator_mapping.remove(&indicator);
-                    }
-                }
-            }
-        }
+        let mut builder = LsColorsBuilder::default();
+        builder.add_from_string(input);
+        builder.build()
     }
 
     /// Get the ANSI style for a given path.
@@ -402,22 +439,20 @@ impl LsColors {
         if indicator == Indicator::RegularFile {
             // Note: using '.to_str()' here means that filename
             // matching will not work with invalid-UTF-8 paths.
-            let filename = file.file_name().to_str()?.to_ascii_lowercase();
-
-            // We need to traverse LS_COLORS from back to front
-            // to be consistent with `ls`:
-            for (suffix, style) in self.suffix_mapping.iter().rev() {
-                // Note: For some reason, 'ends_with' is much
-                // slower if we omit `.as_str()` here:
-                if filename.ends_with(suffix.as_str()) {
-                    return style.as_ref();
-                }
+            let filename = file.file_name();
+            if let Some(style) = self.style_for_str(filename.to_str()?) {
+                return Some(style);
             }
         }
 
         self.style_for_indicator(indicator)
     }
 
+    /// Get the ANSI style for a string. This does not have to be a valid filepath.
+    pub fn style_for_str(&self, file_str: &str) -> Option<&Style> {
+        self.suffixes.get(file_str)
+    }
+
     /// Get the ANSI style for a path, given the corresponding `Metadata` struct.
     ///
     /// *Note:* The `Metadata` struct must have been acquired via `Path::symlink_metadata` in
@@ -499,7 +534,13 @@ impl LsColors {
                     _ => indicator,
                 })
             })
-            .or_else(|| self.indicator_mapping.get(&Indicator::Normal))
+            .or_else(|| {
+                if indicator == Indicator::RegularFile && !self.file_normal_fallback {
+                    None
+                } else {
+                    self.indicator_mapping.get(&Indicator::Normal)
+                }
+            })
     }
 }
 
@@ -537,17 +578,34 @@ mod tests {
     fn style_for_path_uses_correct_ordering() {
         let lscolors = LsColors::from_string("*.foo=01;35:*README.foo=33;44");
 
+        // dummy.foo matches to *.foo without getting overriden.
         let style_foo = lscolors.style_for_path("some/folder/dummy.foo").unwrap();
         assert_eq!(FontStyle::bold(), style_foo.font_style);
         assert_eq!(Some(Color::Magenta), style_foo.foreground);
         assert_eq!(None, style_foo.background);
 
+        // README.foo matches to *README.foo by overriding *.foo
         let style_readme = lscolors
             .style_for_path("some/other/folder/README.foo")
             .unwrap();
         assert_eq!(FontStyle::default(), style_readme.font_style);
         assert_eq!(Some(Color::Yellow), style_readme.foreground);
         assert_eq!(Some(Color::Blue), style_readme.background);
+
+        let lscolors = LsColors::from_string("*README.foo=33;44:*.foo=01;35");
+
+        let style_foo = lscolors.style_for_path("some/folder/dummy.foo").unwrap();
+        assert_eq!(FontStyle::bold(), style_foo.font_style);
+        assert_eq!(Some(Color::Magenta), style_foo.foreground);
+        assert_eq!(None, style_foo.background);
+
+        // README.foo matches to *.foo because *.foo overrides *README.foo
+        let style_readme = lscolors
+            .style_for_path("some/other/folder/README.foo")
+            .unwrap();
+        assert_eq!(FontStyle::bold(), style_readme.font_style);
+        assert_eq!(Some(Color::Magenta), style_readme.foreground);
+        assert_eq!(None, style_readme.background);
     }
 
     #[test]
@@ -605,6 +663,22 @@ mod tests {
         }
     }
 
+    #[test]
+    fn style_for_str() {
+        let lscolors = LsColors::from_string("*.wav=00;36:*.rs=1;38;5;202:");
+
+        assert_eq!(lscolors.style_for_str(""), None);
+        assert_eq!(lscolors.style_for_str("test"), None);
+        assert_eq!(
+            lscolors.style_for_str("test.wav").unwrap().foreground,
+            Some(Color::Cyan)
+        );
+        assert_eq!(
+            lscolors.style_for_str("test.rs").unwrap().foreground,
+            Some(Color::Fixed(202))
+        );
+    }
+
     #[test]
     fn style_for_directory() {
         let tmp_dir = temp_dir();
@@ -786,4 +860,98 @@ mod tests {
         let style = lscolors.style_for_path(&tmp_file);
         assert_eq!(None, style);
     }
+
+    #[test]
+    fn file_normal_fallback() {
+        let tmp_dir = temp_dir();
+        let tmp_file_path = create_file(tmp_dir.path().join("test-file"));
+
+        let lscolors = LsColors::from_string("no=01;31");
+        let style = lscolors.style_for_path(&tmp_file_path).unwrap();
+        assert_eq!(Some(Color::Red), style.foreground);
+
+        let lscolors = LsColors::from_string("no=01;31:fi=0");
+        let style = lscolors.style_for_path(&tmp_file_path);
+        assert_eq!(None, style);
+    }
+
+    #[test]
+    fn file_suffix_case() {
+        let assert_bold_fg_magenta = |style: Option<&Style>| {
+            Some(Color::Magenta) == style.and_then(|v| v.foreground)
+                && matches!(style,Some(sty) if sty.font_style.bold)
+        };
+        let assert_bold_fg_green = |style: Option<&Style>| {
+            Some(Color::Green) == style.and_then(|v| v.foreground)
+                && matches!(style,Some(sty) if sty.font_style.bold)
+        };
+
+        // *.jpg is specified only once so any suffix that has .jpg should match
+        // without caring about the letter case
+        let lscolors = LsColors::from_string("*.jpg=01;35:*.Z=01;31");
+        let lowercase_suffix = lscolors.style_for_str("img1.jpg");
+        assert!(assert_bold_fg_magenta(lowercase_suffix));
+        let uppercase_suffix = lscolors.style_for_str("img1.JPG");
+        assert!(assert_bold_fg_magenta(uppercase_suffix));
+        let mixedcase_suffix = lscolors.style_for_str("img1.JpG");
+        assert!(assert_bold_fg_magenta(mixedcase_suffix));
+
+        // *.jpg is specified more than once with different cases and style, so
+        // case should matter here
+        let lscolors = LsColors::from_string("*.jpg=01;35:*.JPG=01;32");
+        let lowercase_suffix = lscolors.style_for_str("img1.jpg");
+        assert!(assert_bold_fg_magenta(lowercase_suffix));
+        let uppercase_suffix = lscolors.style_for_str("img1.JPG");
+        assert!(assert_bold_fg_green(uppercase_suffix));
+        let mixedcase_suffix = lscolors.style_for_str("img1.JpG");
+        assert!(mixedcase_suffix.is_none());
+
+        // *.jpg is specified more than once with different cases but style is same, so
+        // case can ignored
+        let lscolors = LsColors::from_string("*.jpg=01;35:*.JPG=01;35");
+        let lowercase_suffix = lscolors.style_for_str("img1.jpg");
+        assert!(assert_bold_fg_magenta(lowercase_suffix));
+        let uppercase_suffix = lscolors.style_for_str("img1.JPG");
+        assert!(assert_bold_fg_magenta(uppercase_suffix));
+        let mixedcase_suffix = lscolors.style_for_str("img1.JpG");
+        assert!(assert_bold_fg_magenta(mixedcase_suffix));
+
+        // last *.jpg gets more priority resulting in same style across
+        // different cases specified, so case can ignored
+        let lscolors = LsColors::from_string("*.jpg=01;32:*.jpg=01;35:*.JPG=01;35");
+        let lowercase_suffix = lscolors.style_for_str("img1.jpg");
+        assert!(assert_bold_fg_magenta(lowercase_suffix));
+        let uppercase_suffix = lscolors.style_for_str("img1.JPG");
+        assert!(assert_bold_fg_magenta(uppercase_suffix));
+        let mixedcase_suffix = lscolors.style_for_str("img1.JpG");
+        assert!(assert_bold_fg_magenta(mixedcase_suffix));
+
+        // same as above with different combinations
+        let lscolors = LsColors::from_string("*.jpg=01;32:*.jpg=01;35:*.JPG=01;32:*.jpg=01;32");
+        let lowercase_suffix = lscolors.style_for_str("img1.jpg");
+        assert!(assert_bold_fg_green(lowercase_suffix));
+        let uppercase_suffix = lscolors.style_for_str("img1.JPG");
+        assert!(assert_bold_fg_green(uppercase_suffix));
+        let mixedcase_suffix = lscolors.style_for_str("img1.JpG");
+        assert!(assert_bold_fg_green(mixedcase_suffix));
+
+        // last *.jpg gets more priority resulting in different style across
+        // different cases specified, so case matters
+        let lscolors = LsColors::from_string("*.jpg=01;32:*.jpg=01;35:*.JPG=01;32");
+        let lowercase_suffix = lscolors.style_for_str("img1.jpg");
+        assert!(assert_bold_fg_magenta(lowercase_suffix));
+        let uppercase_suffix = lscolors.style_for_str("img1.JPG");
+        assert!(assert_bold_fg_green(uppercase_suffix));
+        let mixedcase_suffix = lscolors.style_for_str("img1.JpG");
+        assert!(mixedcase_suffix.is_none());
+
+        // same as above with different combinations
+        let lscolors = LsColors::from_string("*.jpg=01;32:*.jpg=01;35:*.JPG=01;35:*.jpg=01;32");
+        let lowercase_suffix = lscolors.style_for_str("img1.jpg");
+        assert!(assert_bold_fg_green(lowercase_suffix));
+        let uppercase_suffix = lscolors.style_for_str("img1.JPG");
+        assert!(assert_bold_fg_magenta(uppercase_suffix));
+        let mixedcase_suffix = lscolors.style_for_str("img1.JpG");
+        assert!(mixedcase_suffix.is_none());
+    }
 }
